{% load i18n verbose_name_tags %}
<div class="box warning cf">
    <strong>{% trans "Please review the following warnings:" %}</strong>

    <ul>
    {% for form in forms %}
        {% for warning in form.warnings %}
            <li>{{ warning }}</li>
        {% endfor %}
        {% if form.warnings %}
            <label for="{{ form.ignore_warnings.auto_id }}">
                {{ form.ignore_warnings }} {{ form.ignore_warnings.label }}
            </label>
        {% endif %}
    {% endfor %}

    {% for formset in formsets %}
        {% for form in formset.forms %}
            {% for warning in form.warnings %}
              <li>{{ warning }}</li>
            {% endfor %}
            {% if form.warnings %}
                <label for="{{ form.ignore_warnings.auto_id }}">
                    {{ form.ignore_warnings }} {{ form.ignore_warnings.label }}
                </label>
            {% endif %}
        {% endfor %}
    {% endfor %}
    </ul>
</div>
